package com.inspect;
import java.lang.reflect.Constructor;
import java.lang.reflect.Executable;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Parameter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.stream.Stream;

import com.nodes.Node;

/**
 * CPSC 501
 * Inspector starter class
 *
 * @author Jonathan Hudson
 */
public class Inspector {

    public void inspect(Object obj, boolean recursive) {
        Class<?> c = obj.getClass();
        
        try {
            inspectClass(c, obj, recursive, 0);
        } catch (IllegalArgumentException | IllegalAccessException | NoSuchFieldException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    private void inspectClass(Class<?> c, Object obj, boolean recursive, int depth) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
        Node root = Node.root();
        properInspection(root, obj, c, recursive, depth);
        root.render(true);
    }


  
    private String[] modifiersToList(String mods) {
        if(mods.length() == 0) {
            return new String[] {};
        }
        return mods.split(" ");
    }
    
    private void inspectConstructorExceptions(Node source, Constructor<?> current, boolean recursive, int depth) {
    	Class<?>[] exceptons = current.getExceptionTypes();
    	Arrays.sort(exceptons, (a, b) -> a.getName().compareTo(b.getName()));
        if(exceptons.length > 0) {
        	source.addListNode("Exceptions", Stream.of(exceptons).map((a) -> a.getName()).toArray(String[]::new));
//            Node exceptions = source.addEntrypoint("Exceptions");
//            for(Class<?> e : exceptons) {
//                exceptions.addInfoNode("Name", e.getName());
//            }
        }
    }
    
    private void inspectParameters(Node source, Executable current, boolean recursive, int depth) throws IllegalAccessException, NoSuchFieldException {
    	Parameter[] ptypes = current.getParameters();
    	Arrays.sort(ptypes, (a, b) -> a.getName().compareTo(b.getName()));
        if(ptypes.length > 0) {
        	source.addListNode("Parameter Types", Stream.of(ptypes).map((a) -> a.getType().getName()).toArray(String[]::new));

            
        }
    }

    private void inspectConstructor(Node source, Constructor<?> current, boolean recursive, int depth) throws IllegalAccessException, NoSuchFieldException {
        
        Node subConstruct = source.addListEntrypoint();
        subConstruct.addInfoNode("Name", current.getName());

        inspectConstructorExceptions(subConstruct, current, recursive, depth);
        

        inspectParameters(subConstruct, current, recursive, depth);



        subConstruct.addListNode("Modifiers", modifiersToList(Modifier.toString(current.getModifiers())));
    }

    private void inspectConstructorsAll(Node source, Class<?> current, boolean recursive, int depth) throws IllegalAccessException, NoSuchFieldException {
        Constructor<?>[] constructors = current.getDeclaredConstructors();
        Arrays.sort(constructors, (a, b) -> a.getName().compareTo(b.getName()));
        if(constructors.length > 0) {
            Node constructorsNode = source.addEntrypoint("Constructors");
            for(Constructor<?> a : constructors) {
                inspectConstructor(constructorsNode, a, recursive, depth);
            }
        }
    }

    private void inspectMethod(Node source, Method method, boolean recursive, int depth) throws IllegalAccessException, NoSuchFieldException {
        Node methodNode = source.addListEntrypoint();
        methodNode.addInfoNode("Name", method.getName());
        methodNode.addListNode("Exceptions", Stream.of(method.getExceptionTypes()).map((a) -> a.getName()).toArray(String[]::new));
        methodNode.addListNode("Parameter Types", Stream.of(method.getParameterTypes()).map((a) -> a.getName()).toArray(String[]::new));
        methodNode.addInfoNode("Return Type", method.getReturnType().getName());
        methodNode.addListNode("Modifiers", modifiersToList(Modifier.toString(method.getModifiers())));
    }

    private void inspectMethodsAll(Node source, Class<?> current, boolean recursive, int depth) throws IllegalAccessException, NoSuchFieldException {
        Method[] methods = current.getDeclaredMethods();
        Arrays.sort(methods, (a, b) -> a.getName().compareTo(b.getName()));
//        System.out.println("Hello: " + methods[0].);
        if(methods.length > 0) {
            Node methodsNode = source.addEntrypoint("Methods");
            for(Method method : methods) {
                inspectMethod(methodsNode, method, recursive, depth);
            }
        }
    }



    private void handlePrimitiveArrayField(Node source, Object instance, Field field, boolean recurse, int depth) throws IllegalAccessException {
        assert field.getType().isArray(); // Make sure the type is actually an array.
        assert field.getType().getComponentType().isPrimitive(); // Make sure the component type is primitive.

        Object obj = field.get(instance);

        ArrayList<String> values = new ArrayList<>();
        if(obj instanceof int[]) {
            for(int s : (int[]) obj) values.add(s + "");
        } else if(obj instanceof float[]) {
            for(float s : (float[]) obj) values.add(s + "");
        } else if(obj instanceof double[]) {
            for(double s : (double[]) obj) values.add(s + "");
        } else if(obj instanceof boolean[]) {
            for(boolean s : (boolean[]) obj) values.add(s + "");
        } else if(obj instanceof byte[]) {
            for(byte s : (byte[]) obj) values.add(s + "");
        } else if(obj instanceof char[]) {
            for(char s : (char[]) obj) values.add(s + "");
        } else if(obj instanceof long[]) {
            for(long s : (long[]) obj) values.add(s + "");
        } else if(obj instanceof short[]) {
            for(short s : (short[]) obj) values.add(s + "");
        } else {
            // THROW EXCEPTION
        }


        
        source.addInfoNode("Length", values.size() + "");
        source.addListNode("Values", values.toArray(String[]::new));

        

    }

    private void handleArrayField(Node source, Object instance, Field field, boolean recurse, int depth) throws IllegalAccessException, NoSuchFieldException {
        assert field.getType().isArray(); // Make sure it actually is an array.

        source.addInfoNode("Component Type", field.getType().getComponentType().getName());
        
       

        if(field.getType().getComponentType().isPrimitive()) {
            // Primitive array.
            handlePrimitiveArrayField(source, instance, field, recurse, depth);
        } else {
        	Object[] obj = (Object[]) field.get(instance);
        	if(!recurse) {
        		
                String[] arr = Stream.of(obj).map((a) -> renderFieldValue(a)).toArray(String[]::new);
                source.addListNode("Values", arr);

        	} else {
        		Node list = source.addEntrypoint("Values");
        		for(Object o : obj) {
        			if(o == null) {
        				list.addInfoNode("-", "null");
        			} else {
        				Node sub = list.addEntrypoint(o.toString());
                		properInspection(sub, o, o.getClass(), recurse, depth);
        			}
        			
        		}
        		
        		
        		
        	}
            
        }

    }

    private String renderFieldValue(Object o) {
    	if(o == null) {
    		return "null";
    	} else {
    		return o.toString();
    	}
    }
    
    private void inspectField(Node source, Object instance, Field field, boolean recursive, int depth) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
        Node methodNode = source.addListEntrypoint();
        methodNode.addInfoNode("Name", field.getName());
        methodNode.addInfoNode("Type", field.getType().getCanonicalName());
        methodNode.addListNode("Modifiers", modifiersToList(Modifier.toString(field.getModifiers())));

        field.setAccessible(true);
        methodNode.setParentName(field.getType().getName());

        if(field.getType().isArray()) {
            // We have special handlers for arrays.
            handleArrayField(methodNode, instance, field, recursive, depth);
          
        } else {
            if(!recursive) {
                // If it is recursive we will not even try to explore references!
                methodNode.addInfoNode("Value", renderFieldValue(field.get(instance)));
            } else {
                if(field.getType().isPrimitive()) {
                    // It is a primitive anyways
                    methodNode.addInfoNode("Value", renderFieldValue(field.get(instance)));
                } else {
                    // We need to recurse on the object.
                	Object subInstance = field.get(instance);
                	if(subInstance == null) {
                		methodNode.addInfoNode("Value", "null");
                	} else {
                		Node value = methodNode.addEntrypoint("Value");
                        properInspection(value, subInstance, subInstance.getClass(), recursive, depth);
                	}
                	
                    
                }
            }
        }
       
    }

    private void inspectFieldsAll(Node source, Object instance, Class<?> current, boolean recursive, int depth) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
        Field[] fields = current.getDeclaredFields();
        Arrays.sort(fields, (a, b) -> a.getName().compareTo(b.getName()));
        if(fields.length > 0) {
            Node fieldsNode = source.addEntrypoint("Fields");
            for(Field field : fields) {
                inspectField(fieldsNode, instance, field, recursive, depth);
            }
        }
    }
    
    private void inspectInterface(Node source, Object instance, Class<?> current, boolean recursive, int depth) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
    	Node subINode = source.addListEntrypoint();
        properInspection(subINode, instance, current, recursive, depth);
    }
    
    private void inspectInterfacesAll(Node source, Object instance, Class<?> current, boolean recursive, int depth) throws IllegalAccessException, IllegalAccessException, NoSuchFieldException {
    	Class<?>[] interfaces = current.getInterfaces();
        if(interfaces.length > 0) {
            Node interfacesNode = source.addEntrypoint("Interfaces");
        

            for(Class<?> a : interfaces) {
                inspectInterface(interfacesNode, instance, a, recursive, depth);
            }

        }
    }

    public void properInspection(Node source, Object instance, Class<?> current, boolean recursive, int depth) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
        
        source.addInfoNode("Class", current.getName());
        source.setParentName(current.getName());
        
        // Explore the superclass.
        if(current.getSuperclass() != null) {
            Node superClassNode = source.addEntrypoint("Superclass");
            properInspection(superClassNode, instance, current.getSuperclass(), recursive, depth);
        }

        // Explore all the interfaces.
        inspectInterfacesAll(source, instance, current, recursive, depth);

        // Explore all the constructors.
        inspectConstructorsAll(source, current, recursive, depth);
        
        // Explore all the methods.
        inspectMethodsAll(source, current, recursive, depth);

        // Explore all  the fields.
        inspectFieldsAll(source, instance, current, recursive, depth);

    }

}