package com.nodes;

import java.util.ArrayList;
import java.util.Optional;

/**
 * Represents a node in the tree structure.
 */
public class Node {
	// The depth of the node within the tree.
    int depth;
    // The title of the node, may be null.
    String title;
    
    // Parentname
    String parentName;

    // The children of this node.
    ArrayList<Node> children = new ArrayList<>();
    
    // The names of the entrypoints (if they exist)
    ArrayList<String> eps = new ArrayList<>();
    
    boolean isListed;


    /**
     * Creates a new node from depth, title, and if it is listed.
     * 
     * @param depth: the depth of the node.
     * @param title: the title of the node.
     * @param isListed: if the node is listed.
     */
    protected Node(int depth, String title, boolean isListed) {
        this.depth = depth;
        this.title = title;
        this.isListed = isListed;
    }
    
    /**
     * Sets the tree parent name.
     * @param name: name to set to.
     */
    public void setParentName(String name) {
    	this.parentName = name;
    }

 
    /**
     * Creates a root node.
     * @return The root node.
     */
    public static Node root() {
        return new Node(0, null, false);
    }

    /**
     * Adds a list entrypoint to the tree.
     * @return an entrypoint node.
     */
    public EntrypointNode addListEntrypoint() {
        EntrypointNode newNode = new EntrypointNode(this.depth + 1);
        this.children.add(newNode);
        this.eps.add(null);
        newNode.setParentName(this.parentName);
        return newNode;
    }

    /**
     * Adds an entrypoint node to the tree.
     * 
     * @param entrypoint: the name of the entrypoint
     * @return the entrypoint node.
     */
    public Node addEntrypoint(String entrypoint) {
        Node newNode = new Node(this.depth + 1, null, false);
        this.children.add(newNode);
        this.eps.add(entrypoint);
        newNode.setParentName(this.parentName);
        return newNode;
    }

    /**
     * Adds an info node to the tree.
     * 
     * @param title: the title (key) of the info node.
     * @param datum: the value of the info node.
     * @return the info node.
     */
    public InfoNode addInfoNode(String title, String datum) {
        InfoNode newNode = new InfoNode(this.depth + 1, title, datum);
        this.children.add(newNode);
        this.eps.add(null);
        newNode.setParentName(this.parentName);
        return newNode;
    }
    
    /**
     * Looks up an entrypoint node.
     * 
     * @param name: the name of the nods.
     * @return the entrypoint
     */
    public Optional<Node> lookupEntrypoint(String name) {
    	int index = this.eps.indexOf(name);
    	if(index == -1) {
    		return Optional.empty();
    	} else {
    		return Optional.of(this.children.get(index));
    	}
    	
    }
    
    /**
     * Gets a child from the children nods.
     * 
     * @param position: the position at which to grab he node.
     * @return the node a the position
     */
    public Optional<Node> getChild(int position) {
    	if(this.children.size() <= position) {
    		return Optional.empty();
    	} else {
    		return Optional.of(this.children.get(position));
    	}
    }
    
    /**
     * Looks up a KV node.
     * 
     * @param key: the key to lookup wih
     * @return the value
     */
    public Optional<String> lookupKey(String key) {
    	for(Node n : children) {
    		if(n instanceof InfoNode && ((InfoNode) n).title == key) {
    			return Optional.of(((InfoNode) n).datum);
    		}
    	}
    	return Optional.empty();
    }
    
    /**
     * Looks up a KV list node.
     * 
     * @param key: the key to lookup wih
     * @return the value
     */
    public Optional<String[]> lookupKeyList(String key) {
    	for(Node n : children) {
    		if(n instanceof ListNode && ((ListNode) n).title == key) {
    			return Optional.of(((ListNode) n).values);
    		}
    	}
    	return Optional.empty();
    }

    /**
     * Adds a list node to the tree.
     * 
     * @param title: the title of the list.
     * @param vas: the values in the nde.
     * @return The list node.
     */
    public ListNode addListNode(String title, String[] vas) {
        if (vas.length == 0) {
            return null;
        }
        ListNode newNode = new ListNode(this.depth + 1, title, vas);
        this.children.add(newNode);
        this.eps.add(null);
        return newNode;
    }

    
    /**
     * Renders a node.
     * @param tabulate
     */
    public void render(boolean tabulate) {

        for (int i = 0; i < this.children.size(); i++) {
            if (this.eps.get(i) != null) {
                this.tabulate(this.depth);
                System.out.println(this.eps.get(i) + " (" + this.parentName + ")" + ":");
            }

            this.children.get(i).render(this.isListed ? i > 0 : true);
        }
    }

    /**
     * Renders the title of a node.
     * 
     * @param newLine: should we return a newline
     * @param tabulate: should we tabulate.
     */
    protected void renderTitle(boolean newLine, boolean tabulate) {

        if(tabulate)
            this.tabulate(this.depth - 1);
        if(this.title.equals("-")) {
        	System.out.print(this.title + " ");
        } else {
        	System.out.print(this.title + ": ");
        }
        
        

        if (newLine) {
            System.out.println();
        }
    }

    /**
     * Tabulates out the fields at a certain depth.
     * 
     * @param depth: the depth to tabulate out at
     */
    protected void tabulate(int depth) {
        for (int i = 0; i < depth; i++)
            System.out.print("  ");
       
    }

}
